#include "casm/clexmonte/monte_calculator/KineticCalculator.hh"

#include "casm/casm_io/json/InputParser_impl.hh"
#include "casm/clexmonte/events/io/json/event_data_json_io.hh"
#include "casm/clexmonte/misc/to_json.hh"
#include "casm/clexmonte/monte_calculator/MonteEventData.hh"
#include "casm/clexmonte/monte_calculator/analysis_functions.hh"
#include "casm/clexmonte/monte_calculator/kinetic_events.hh"
#include "casm/clexmonte/monte_calculator/kinetic_sampling_functions.hh"
#include "casm/clexmonte/monte_calculator/modifying_functions.hh"
#include "casm/clexmonte/monte_calculator/sampling_functions.hh"
#include "casm/clexmonte/monte_calculator/selected_event_functions.hh"
#include "casm/clexmonte/run/functions.hh"
#include "casm/configuration/io/json/Configuration_json_io.hh"
#include "casm/monte/events/OccEventProposal.hh"
#include "casm/monte/run_management/RunManager.hh"
#include "casm/monte/sampling/RequestedPrecisionConstructor.hh"
#include "casm/monte/sampling/io/json/SelectedEventFunctions_json_io.hh"

namespace CASM {
namespace clexmonte {
namespace kinetic_2 {

KineticPotential::KineticPotential(std::shared_ptr<StateData> _state_data)
    : BaseMontePotential(_state_data),
      state(*state_data->state),
      formation_energy_clex(
          get_clex(*state_data->system, state, "formation_energy")) {}

/// \brief Calculate (per_supercell) potential value
double KineticPotential::per_supercell() {
  return formation_energy_clex->per_supercell();
}

/// \brief Calculate (per_unitcell) potential value
double KineticPotential::per_unitcell() {
  return formation_energy_clex->per_unitcell();
}

/// \brief Calculate change in (per_supercell) potential value due
///     to a series of occupation changes
double KineticPotential::occ_delta_per_supercell(
    std::vector<Index> const &linear_site_index,
    std::vector<int> const &new_occ) {
  return formation_energy_clex->occ_delta_value(linear_site_index, new_occ);
}

KineticCalculator::KineticCalculator()
    : BaseMonteCalculator(
          "KineticCalculator",   // calculator_name
          {},                    // required_basis_set,
          {},                    // required_local_basis_set,
          {"formation_energy"},  // required_clex,
          {},                    // required_multiclex,
          {},                    // required_local_clex,
          {},                    // required_local_multiclex,
          {},                    // required_dof_spaces,
          {},                    // required_params,
          {"verbosity", "mol_composition_tol", "event_data_type",
           "event_selector_type", "impact_table_type",
           "allow_events_with_no_barrier", "assign_allowed_events_only",
           "selected_event_data"},  // optional_params,
          true,                     // time_sampling_allowed,
          true,                     // update_atoms,
          false,                    // save_atom_info,
          false                     // is_multistate_method,
      ) {
  // this could go into base constructor
  this->selected_event = std::make_shared<SelectedEvent>();
  this->selected_event_functions =
      std::make_shared<monte::SelectedEventFunctions>();
  this->selected_event_data = std::make_shared<monte::SelectedEventData>();
}

/// \brief Construct functions that may be used to sample various quantities
///     of the Monte Carlo calculation as it runs
std::map<std::string, state_sampling_function_type>
KineticCalculator::standard_sampling_functions(
    std::shared_ptr<MonteCalculator> const &calculation) const {
  using namespace monte_calculator;

  std::vector<state_sampling_function_type> functions =
      common_sampling_functions(
          calculation, "potential_energy",
          "Potential energy of the state (normalized per primitive cell)");

  // Specific to kmc
  functions.push_back(make_mean_R_squared_collective_isotropic_f(calculation));
  functions.push_back(
      make_mean_R_squared_collective_anisotropic_f(calculation));
  functions.push_back(make_mean_R_squared_individual_isotropic_f(calculation));
  functions.push_back(
      make_mean_R_squared_individual_anisotropic_f(calculation));
  functions.push_back(make_L_isotropic_f(calculation));
  functions.push_back(make_L_anisotropic_f(calculation));
  functions.push_back(make_D_tracer_isotropic_f(calculation));
  functions.push_back(make_D_tracer_anisotropic_f(calculation));
  functions.push_back(make_jumps_per_atom_by_type_f(calculation));
  functions.push_back(make_jumps_per_event_by_type_f(calculation));
  functions.push_back(make_jumps_per_atom_per_event_by_type_f(calculation));
  functions.push_back(make_selected_event_count_by_type_f(calculation));
  functions.push_back(make_selected_event_fraction_by_type_f(calculation));
  functions.push_back(
      make_selected_event_count_by_equivalent_index_f(calculation));
  functions.push_back(
      make_selected_event_fraction_by_equivalent_index_f(calculation));
  functions.push_back(
      make_selected_event_count_by_equivalent_index_and_direction_f(
          calculation));
  functions.push_back(
      make_selected_event_fraction_by_equivalent_index_and_direction_f(
          calculation));

  for (auto f : make_selected_event_count_by_equivalent_index_per_event_type_f(
           calculation)) {
    functions.push_back(f);
  }
  for (auto f :
       make_selected_event_fraction_by_equivalent_index_per_event_type_f(
           calculation)) {
    functions.push_back(f);
  }

  std::map<std::string, state_sampling_function_type> function_map;
  for (auto const &f : functions) {
    function_map.emplace(f.name, f);
  }
  return function_map;
}

/// \brief Construct functions that may be used to sample various quantities
///     of the Monte Carlo calculation as it runs
std::map<std::string, json_state_sampling_function_type>
KineticCalculator::standard_json_sampling_functions(
    std::shared_ptr<MonteCalculator> const &calculation) const {
  std::vector<json_state_sampling_function_type> functions =
      monte_calculator::common_json_sampling_functions(calculation);

  std::map<std::string, json_state_sampling_function_type> function_map;
  for (auto const &f : functions) {
    function_map.emplace(f.name, f);
  }
  return function_map;
}

/// \brief Construct functions that may be used to analyze Monte Carlo
///     calculation results
std::map<std::string, results_analysis_function_type>
KineticCalculator::standard_analysis_functions(
    std::shared_ptr<MonteCalculator> const &calculation) const {
  std::vector<results_analysis_function_type> functions = {
      monte_calculator::make_heat_capacity_f(calculation)};

  std::map<std::string, results_analysis_function_type> function_map;
  for (auto const &f : functions) {
    function_map.emplace(f.name, f);
  }
  return function_map;
}

/// \brief Construct functions that may be used to modify states
StateModifyingFunctionMap KineticCalculator::standard_modifying_functions(
    std::shared_ptr<MonteCalculator> const &calculation) const {
  std::vector<StateModifyingFunction> functions = {
      monte_calculator::make_match_composition_f(calculation),
      monte_calculator::make_enforce_composition_f(calculation)};

  StateModifyingFunctionMap function_map;
  for (auto const &f : functions) {
    function_map.emplace(f.name, f);
  }
  return function_map;
}

/// \brief Construct functions that may be used to collect selected event data
std::optional<monte::SelectedEventFunctions>
KineticCalculator::standard_selected_event_functions(
    std::shared_ptr<MonteCalculator> const &calculation) const {
  using namespace monte_calculator;
  monte::SelectedEventFunctions functions;

  // Event type data:
  functions.insert(make_selected_event_by_type_f(calculation));
  functions.insert(make_selected_event_by_equivalent_index_f(calculation));
  functions.insert(
      make_selected_event_by_equivalent_index_and_direction_f(calculation));
  for (auto f :
       make_selected_event_by_equivalent_index_per_event_type_f(calculation)) {
    functions.insert(f);
  }
  for (auto f : make_local_orbit_composition_f(calculation)) {
    functions.insert(f);
  }

  // Event state data:
  functions.insert(make_dE_activated_by_type_f(calculation));
  functions.insert(make_dE_activated_by_equivalent_index_f(calculation));

  return functions;
}

/// \brief Construct default SamplingFixtureParams
sampling_fixture_params_type
KineticCalculator::make_default_sampling_fixture_params(
    std::shared_ptr<MonteCalculator> const &calculation, std::string label,
    bool write_results, bool write_trajectory, bool write_observations,
    bool write_status, std::optional<std::string> output_dir,
    std::optional<std::string> log_file, double log_frequency_in_s) const {
  monte::SamplingParams sampling_params;
  {
    auto &s = sampling_params;
    s.sampler_names = {"clex.formation_energy",
                       "potential_energy",
                       "mol_composition",
                       "param_composition",
                       "mean_R_squared_collective_isotropic",
                       "mean_R_squared_individual_isotropic",
                       "L_isotropic",
                       "D_tracer_isotropic",
                       "mean_R_squared_collective_anisotropic",
                       "mean_R_squared_individual_anisotropic",
                       "L_anisotropic",
                       "D_tracer_anisotropic",
                       "jumps_per_atom_by_type",
                       "jumps_per_event_by_type",
                       "jumps_per_atom_per_event_by_type"};
    s.do_sample_time = true;
    std::string prefix;
    prefix = "order_parameter.";
    for (auto const &pair : calculation->system()->dof_spaces) {
      s.sampler_names.push_back(prefix + pair.first);
    }
    prefix = "order_parameter.";
    std::string suffix = ".subspace_magnitudes";
    for (auto const &pair : calculation->system()->dof_subspaces) {
      s.sampler_names.push_back(prefix + pair.first + suffix);
    }
    if (write_trajectory) {
      s.do_sample_trajectory = true;
    }
  }

  monte::CompletionCheckParams<statistics_type> completion_check_params;
  {
    auto &c = completion_check_params;
    c.equilibration_check_f = monte::default_equilibration_check;
    c.calc_statistics_f =
        monte::default_statistics_calculator<statistics_type>();

    completion_check_params.cutoff_params.max_count = 100;
  }

  std::vector<std::string> analysis_names = {"heat_capacity"};

  return clexmonte::make_sampling_fixture_params(
      label, calculation->sampling_functions,
      calculation->json_sampling_functions, calculation->analysis_functions,
      sampling_params, completion_check_params, analysis_names, write_results,
      write_trajectory, write_observations, write_status, output_dir, log_file,
      log_frequency_in_s);
}

/// \brief Validate the state's configuration
///
/// Notes:
/// - All configurations are valid (validate_state checks for consistency
///   with the composition conditions)
Validator KineticCalculator::validate_configuration(state_type &state) const {
  return Validator{};
}

/// \brief Validate state's conditions
///
/// Notes:
/// - requires scalar temperature
/// - validate composition consistency
/// - warnings if other conditions are present
Validator KineticCalculator::validate_conditions(state_type &state) const {
  // Validate system
  if (this->system == nullptr) {
    throw std::runtime_error(
        "Error in KineticCalculator::validate_conditions: system==nullptr");
  }

  // validate state.conditions
  monte::ValueMap const &conditions = state.conditions;
  Validator v;
  v.insert(validate_keys(conditions.scalar_values, {"temperature"} /*required*/,
                         {} /*optional*/, "scalar", "condition",
                         false /*throw_if_invalid*/));
  v.insert(validate_keys(conditions.vector_values, {} /*required*/,
                         {"param_composition", "mol_composition"} /*optional*/,
                         "vector", "condition", false /*throw_if_invalid*/));
  v.insert(validate_composition_consistency(
      state, get_composition_converter(*this->system),
      this->mol_composition_tol));
  return v;
}

/// \brief Validate state
Validator KineticCalculator::validate_state(state_type &state) const {
  Validator v;
  v.insert(this->validate_configuration(state));
  v.insert(this->validate_conditions(state));
  if (!v.valid()) {
    return v;
  }

  // check if configuration is consistent with conditions
  auto const &composition_calculator =
      get_composition_calculator(*this->system);
  auto const &composition_converter = get_composition_converter(*this->system);

  Eigen::VectorXd mol_composition =
      composition_calculator.mean_num_each_component(get_occupation(state));
  Eigen::VectorXd param_composition =
      composition_converter.param_composition(mol_composition);

  Eigen::VectorXd target_mol_composition =
      get_mol_composition(*this->system, state.conditions);
  Eigen::VectorXd target_param_composition =
      composition_converter.param_composition(target_mol_composition);

  if (!CASM::almost_equal(mol_composition, target_mol_composition,
                          this->mol_composition_tol)) {
    std::stringstream msg;
    msg << "***" << std::endl;
    msg << "Calculated composition is not consistent with conditions "
           "composition."
        << std::endl;
    msg << "Calculated composition:" << std::endl;
    msg << "- mol_composition: " << mol_composition.transpose() << std::endl;
    msg << "- param_composition: " << param_composition.transpose()
        << std::endl;
    msg << "Conditions:" << std::endl;
    msg << "- mol_composition: " << target_mol_composition.transpose()
        << std::endl;
    msg << "- param_composition: " << target_param_composition.transpose()
        << std::endl;
    msg << "***" << std::endl;
    v.error.insert(msg.str());
  }
  return v;
}

/// \brief Validate and set the current state, construct state_data, construct
///     potential
///
/// \param state State to set
/// \param occ_location Pointer to OccLocation to use, or may be nullptr
void KineticCalculator::set_state_and_potential(
    state_type &state, monte::OccLocation *occ_location) {
  // Validate system
  if (this->system == nullptr) {
    throw std::runtime_error(
        "Error in KineticCalculator::run: system==nullptr");
  }

  // Validate state
  Validator v = this->validate_state(state);
  clexmonte::print(CASM::log(), v);
  if (!v.valid()) {
    throw std::runtime_error(
        "Error in KineticCalculator::run: Invalid initial state");
  }

  // Make state data
  this->state_data =
      std::make_shared<StateData>(this->system, &state, occ_location);

  // Make potential calculator
  this->potential = std::make_shared<KineticPotential>(this->state_data);
}

/// \brief Set event data (includes calculating all rates), using current
/// state data
///
/// Notes:
/// - Validates this->state_data is not null
/// - Validates this->state_data->occ_location is not null
///
/// \param engine The random number generator engine used to select events
///     and timesteps. If nullptr, a new engine is constructed,
///     seeded by std::random_device
void KineticCalculator::set_event_data(std::shared_ptr<engine_type> engine) {
  if (this->state_data == nullptr) {
    throw std::runtime_error(
        "Error in KineticCalculator::set_event_data: "
        "this->state_data==nullptr");
  }
  if (this->state_data->occ_location == nullptr) {
    throw std::runtime_error(
        "Error in KineticCalculator::set_event_data: "
        "this->state_data->occ_location==nullptr");
  }
  if (this->event_data == nullptr) {
    throw std::runtime_error(
        "Error in KineticCalculator::set_event_data: "
        "this->event_data==nullptr");
  }

  // Currently, event_filters are only set at _reset() by reading from params
  this->event_data->update(this->state_data, event_filters, engine);
}

/// \brief Perform a single run, evolving current state
void KineticCalculator::run(state_type &state, monte::OccLocation &occ_location,
                            run_manager_type<engine_type> &run_manager) {
  if (run_manager.sampling_fixtures.size() == 0) {
    throw std::runtime_error(
        "Error in KineticCalculator::run: "
        "run_manager.sampling_fixtures.size()==0");
  }

  // Set state and potential
  // - Validates this->system is not null
  // - Validates state
  // - Throw if validation fails
  // - Constructs this->state_data
  // - Constructs this->potential
  this->set_state_and_potential(state, &occ_location);

  // Set event data
  // - Set or re-set state in event_state_calculators
  // - Update event_data if supercell has changed
  // - Throw if this->state_data is null
  // - Constructs this->event_data->event_selector
  // - Calculates all rates
  std::cout << "Setting event data ... " << std::endl;
  this->set_event_data(run_manager.engine);
  std::cout << "Setting event data ... DONE" << std::endl << std::endl;

  // Construct EventDataSummary
  std::cout << "Generating event data summary ... " << std::endl;
  MonteEventData monte_event_data(this->event_data, nullptr);
  double energy_bin_width = 0.1;
  double freq_bin_width = 0.1;
  double rate_bin_width = 0.1;

  EventDataSummary event_data_summary(this->state_data, monte_event_data,
                                      energy_bin_width, freq_bin_width,
                                      rate_bin_width);
  std::cout << "Generating event data summary ... DONE" << std::endl
            << std::endl;
  print(std::cout, event_data_summary);

  if (event_data_summary.n_events_allowed == 0) {
    throw std::runtime_error("Error: Cannot run. No allowed events.");
  }

  // Construct KMCData
  this->kmc_data = std::make_shared<kmc_data_type>();

  // Update atom_name_index_list
  // -- These do not change (no atoms moving to/from reservoirs) --
  auto event_system = get_event_system(*this->system);

  // Optional: Manages constructing histogram data structures
  // and collecting selected event data

  std::optional<monte::SelectedEventDataCollector> collector;
  if (this->selected_event_function_params) {
    if (this->selected_event_functions == nullptr) {
      throw std::runtime_error(
          "Error in KineticCalculator::run: "
          "this->selected_event_functions==nullptr");
    }
    collector = monte::SelectedEventDataCollector(
        *selected_event_functions, *selected_event_function_params,
        selected_event_data);
  }

  // Check this->selected_event is not null
  if (this->selected_event == nullptr) {
    throw std::runtime_error(
        "Error in KineticCalculator::run: this->selected_event==nullptr");
  }

  this->event_data->run(state, occ_location, *this->kmc_data,
                        *this->selected_event, collector, run_manager,
                        event_system);

  //  // Function to set selected event
  //  bool requires_event_state =
  //      collector.has_value() && collector->requires_event_state;
  //  auto set_selected_event_f = [=](SelectedEvent &selected_event) {
  //    this->event_data->select_event(selected_event, requires_event_state);
  //  };
  //
  //  // Run Kinetic Monte Carlo at a single condition
  //  kinetic_monte_carlo_v2<EventID>(state, occ_location, *this->kmc_data,
  //                                  *this->selected_event,
  //                                  set_selected_event_f, collector,
  //                                  run_manager, event_system);

  // Warn if events with no barrier were encountered
  check_n_not_normal(this->event_data->n_not_normal());
}

/// \brief Print a warning to std::cerr if events with no barrier were
///     encountered
void KineticCalculator::check_n_not_normal(
    std::map<std::string, Index> const &n_not_normal) const {
  if (n_not_normal.empty()) {
    return;
  }

  std::cerr << "## WARNING: EVENTS WITH NO BARRIER ##################\n"
               "#                                                   #\n"
               "# Events with no barrier are treated as having a    #\n"
               "# rate equal to the attempt frequency.              #\n"
               "#                                                   #\n"
               "# Number encountered by type:                       #\n";
  for (auto const &pair : n_not_normal) {
    std::cerr << "  - " << pair.first << ": " << pair.second << "\n";
  }
  std::cerr << "#                                                   #\n"
               "#####################################################\n"
            << std::endl;
}

/// \brief Perform a single run, evolving one or more states
void KineticCalculator::run(int current_state, std::vector<state_type> &states,
                            std::vector<monte::OccLocation> &occ_locations,
                            run_manager_type<engine_type> &run_manager) {
  throw std::runtime_error(
      "Error: KineticCalculator does not allow multi-state runs");
}

template <bool DebugMode>
void KineticCalculator::make_complete_event_data_impl() {
  if constexpr (DebugMode) {
    auto &log = CASM::log();
    log << "!! make_complete_event_data_impl !! " << std::endl;
    log << "!! DebugMode=" << DebugMode << " !! " << std::endl;
    log << std::endl;
  }

  this->event_data = std::make_shared<CompleteKineticEventData<DebugMode>>(
      system, event_filters, allow_events_with_no_barrier);
}

template <bool DebugMode>
void KineticCalculator::make_allowed_event_data_impl() {
  if constexpr (DebugMode) {
    auto &log = CASM::log();
    log << "!! make_allowed_event_data_impl !! " << std::endl;
    log << "!! DebugMode=" << DebugMode << " !! " << std::endl;
    log << std::endl;
  }

  // current testing does not show this to be helpful - fix to false
  bool use_map_index = false;

  if (this->event_selector_type ==
      kinetic_event_selector_type::vector_sum_tree) {
    this->event_data = std::make_shared<AllowedKineticEventData<
        vector_sum_tree_event_selector_type, DebugMode>>(
        system, allow_events_with_no_barrier, use_map_index,
        use_neighborlist_impact_table, assign_allowed_events_only);

  } else if (this->event_selector_type ==
             kinetic_event_selector_type::sum_tree) {
    this->event_data = std::make_shared<
        AllowedKineticEventData<sum_tree_event_selector_type, DebugMode>>(
        system, allow_events_with_no_barrier, use_map_index,
        use_neighborlist_impact_table, assign_allowed_events_only);

  } else if (this->event_selector_type ==
             kinetic_event_selector_type::direct_sum) {
    this->event_data = std::make_shared<
        AllowedKineticEventData<direct_sum_event_selector_type, DebugMode>>(
        system, allow_events_with_no_barrier, use_map_index,
        use_neighborlist_impact_table, assign_allowed_events_only);

  } else {
    throw std::runtime_error(
        "Error in KineticCalculator: "
        "invalid event_selector_type for event_data_type");
  }
}

/// \brief Reset the derived Monte Carlo calculator
///
/// Parameters:
///
///   verbosity: str or int, default=10
///       If integer, the allowed range is `[0,100]`. If string, then:
///       - "none" is equivalent to integer value 0
///       - "quiet" is equivalent to integer value 5
///       - "standard" is equivalent to integer value 10
///       - "verbose" is equivalent to integer value 20
///       - "debug" is equivalent to integer value 100
void KineticCalculator::_reset() {
  // -- Parsing ----------------------------

  // `params` is BaseMonteCalculator::params
  // - originally "calculation_options" from the run params input file
  // - `params` argument of MonteCalculator Python constructor

  ParentInputParser parser{params};

  // "verbosity": str or int, default=10
  this->verbosity_level = parse_verbosity(parser);
  CASM::log().set_verbosity(this->verbosity_level);

  auto &log = CASM::log();
  log.read("KineticCalculator parameters");
  log.indent() << "verbosity=" << this->verbosity_level << std::endl;

  // "mol_composition_tol": float, default=CASM::TOL
  this->mol_composition_tol = CASM::TOL;
  parser.optional(this->mol_composition_tol, "mol_composition_tol");
  log.indent() << "mol_composition_tol=" << this->mol_composition_tol
               << std::endl;

  // TODO: enumeration

  // TODO: Read event_filters from params
  this->event_filters = std::nullopt;
  log.indent() << "event_filters=" << qto_json(this->event_filters)
               << std::endl;

  // Read selected event data params
  this->selected_event_function_params.reset();
  if (parser.self.contains("selected_event_data")) {
    auto selected_event_data_subparser =
        parser.subparse<monte::SelectedEventFunctionParams>(
            "selected_event_data");
    if (selected_event_data_subparser->valid()) {
      this->selected_event_function_params =
          std::move(selected_event_data_subparser->value);
    }
  }
  log.indent() << "selected_event_data=" << qto_json(this->event_filters)
               << std::endl;

  // Read "event_data_type"
  // - "high_memory": complete event list,
  // - "default" (default): allowed event list w/ vector index, or
  // - "low_memory": allowed event list w/ map index
  std::string event_data_type_str = "default";
  parser.optional(event_data_type_str, "event_data_type");
  if (event_data_type_str == "high_memory") {
    this->event_data_type = kinetic_event_data_type::high_memory;
    log.indent() << "event_data_type="
                 << "\"high_memory\"" << std::endl;
  } else if (event_data_type_str == "default") {
    this->event_data_type = kinetic_event_data_type::default_memory;
    log.indent() << "event_data_type="
                 << "\"default\"" << std::endl;

    if (event_filters.has_value()) {
      parser.insert_error(
          "event_data_type",
          "event_filters are not supported by event_data_type 'default'");
    }

  } /* else if (event_data_type_str == "low_memory") { // not currently used
    this->event_data_type = kinetic_event_data_type::low_memory;

    if (event_filters.has_value()) {
      parser.insert_error(
          "event_data_type",
          "event_filters are not supported by event_data_type 'low_memory'");
    }
  } */
  else {
    parser.insert_error("event_data_type",
                        "Invalid event_data_type: " + event_data_type_str);
  }

  // Read "event_selector_type"
  // - "vector_sum_tree" (default): binary sum tree (log complexity) built
  //   using std::vector
  // - "sum_tree": binary sum tree (log complexity) built using std::map and
  //   linked lists
  // - "direct_sum": direct sum of rate vector (linear complexity)
  std::string event_selector_type_str = "vector_sum_tree";
  parser.optional(event_selector_type_str, "event_selector_type");
  if (event_selector_type_str == "vector_sum_tree") {
    this->event_selector_type = kinetic_event_selector_type::vector_sum_tree;
    log.indent() << "event_selector_type="
                 << "\"vector_sum_tree\"" << std::endl;
  } else if (event_selector_type_str == "sum_tree") {
    this->event_selector_type = kinetic_event_selector_type::sum_tree;
    log.indent() << "event_selector_type="
                 << "\"sum_tree\"" << std::endl;
  } else if (event_selector_type_str == "direct_sum") {
    this->event_selector_type = kinetic_event_selector_type::direct_sum;
    log.indent() << "event_selector_type="
                 << "\"direct_sum\"" << std::endl;
  } else {
    parser.insert_error("event_selector_type", "Invalid event_selector_type: " +
                                                   event_selector_type_str);
  }

  /// Read "impact_table_type" (optional, only takes affect for "low_memory")
  this->use_neighborlist_impact_table = true;
  std::string impact_table_type_str = "neighborlist";
  parser.optional(impact_table_type_str, "impact_table_type");
  if (impact_table_type_str == "neighborlist") {
    this->use_neighborlist_impact_table = true;
    log.indent() << "impact_table_type="
                 << "\"neighborlist\"" << std::endl;
  } else if (impact_table_type_str == "relative") {
    this->use_neighborlist_impact_table = false;
    log.indent() << "impact_table_type="
                 << "\"relative\"" << std::endl;
  } else {
    parser.insert_error("impact_table_type",
                        "Invalid impact_table_type: " + impact_table_type_str);
  }

  // Read "allow_events_with_no_barrier"
  // - true or false (default)
  this->allow_events_with_no_barrier = false;
  parser.optional(allow_events_with_no_barrier, "allow_events_with_no_barrier");
  log.indent() << "allow_events_with_no_barrier=" << std::boolalpha
               << this->allow_events_with_no_barrier << std::endl;

  // Read "assign_allowed_events_only"
  //
  /// \brief If true (default) check if potentially impacted events are allowed
  ///     and only assign them to the event list if they are (adds an
  ///     additional check, but may reduce the size of the event list).
  ///     Otherwise, assign all potentially impacted events to the event list
  ///     (whether they are allowed will still be checked during the rate
  ///     calculation).
  this->assign_allowed_events_only = true;
  parser.optional(assign_allowed_events_only, "assign_allowed_events_only");
  log.indent() << "assign_allowed_events_only=" << std::boolalpha
               << this->assign_allowed_events_only << std::endl
               << std::endl;
  log.end_section();

  std::stringstream ss;
  ss << "Error in KineticCalculator: error reading calculation "
        "parameters.";
  std::runtime_error error_if_invalid{ss.str()};
  report_and_throw_if_invalid(parser, CASM::log(), error_if_invalid);

  // -- After parsing ----------------------------

  // Make event data
  if (this->event_data_type == kinetic_event_data_type::high_memory) {
    if (this->event_selector_type != kinetic_event_selector_type::sum_tree) {
      throw std::runtime_error(
          "Error in KineticCalculator: "
          "for event_data_type=\"high_memory\", "
          "only event_selector_type=\"sum_tree\" is allowed.");
    }

    if (this->verbosity_level == 100) {
      const bool DebugMode = true;
      this->make_complete_event_data_impl<DebugMode>();
    } else {
      const bool DebugMode = false;
      this->make_complete_event_data_impl<DebugMode>();
    }
  } else {
    // event_data_type == default_memory or low_memory
    if (this->verbosity_level == 100) {
      const bool DebugMode = true;
      this->make_allowed_event_data_impl<DebugMode>();
    } else {
      const bool DebugMode = false;
      this->make_allowed_event_data_impl<DebugMode>();
    }
  }

  return;
}

/// \brief Clone the KineticCalculator
KineticCalculator *KineticCalculator::_clone() const {
  return new KineticCalculator(*this);
}

}  // namespace kinetic_2
}  // namespace clexmonte
}  // namespace CASM

extern "C" {
/// \brief Returns a clexmonte::BaseMonteCalculator* owning a KineticCalculator
CASM::clexmonte::BaseMonteCalculator *make_KineticCalculator() {
  return new CASM::clexmonte::kinetic_2::KineticCalculator();
}
}
